"""
Speech-to-text based recipes
"""

from .superb_asr import SuperbASR
from .superb_pr import SuperbPR
from .superb_sf import SuperbSF

__all__ = [
    "SuperbASR",
    "SuperbPR",
    "SuperbSF",
]
